package net.sourceforge.jenesis4java.impl;

/**
 * Copyright (C) 2008, 2010 Richard van Nieuwenhoven - ritchie [at] gmx [dot] at
 * Copyright (C) 2000, 2001 Paul Cody Johnston - pcj@inxar.org <br>
 * This file is part of Jenesis4java. Jenesis4java is free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.<br>
 * Jenesis4java is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.<br>
 * You should have received a copy of the GNU Lesser General Public License
 * along with Jenesis4java. If not, see <http://www.gnu.org/licenses/>.
 */
import net.sourceforge.jenesis4java.ArrayType;
import net.sourceforge.jenesis4java.ClassType;
import net.sourceforge.jenesis4java.CodeWriter;
import net.sourceforge.jenesis4java.Import;
import net.sourceforge.jenesis4java.PrimitiveType;
import net.sourceforge.jenesis4java.Type;

/**
 * Standard <code>Type</code> implementations.
 */
public abstract class MType extends MVM.MCodeable {

    private static final String JAVA_LANG_PACKAGE = "java.lang.";

    /* =============================================================== */
    /* ARRAY TYPE */
    /* =============================================================== */
    static class MArrayType extends MType implements ArrayType {

        Type type;

        int dims;

        public MArrayType(MVM vm, Type type, int dims) {
            super(vm, Type.ARRAY, type.toString());
            this.type = type;
            this.dims = dims;
        }

        public Type getComponentType() {
            return this.type;
        }

        public int getDims() {
            return this.dims;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            out.queue(this.comment);
            this.type.toCode(out);
            for (int i = 0; i < this.dims; i++) {
                out.write("[]");
            }
            return out;
        }
    }

    /* =============================================================== */
    /* CLASS TYPE */
    /* =============================================================== */
    static class MClassType extends MType implements ClassType {

        MClassType(MVM vm, String name) {
            super(vm, Type.CLASS, name);
        }

        public String getName() {
            return this.name;
        }
    }

    /* =============================================================== */
    /* PRIMITIVE TYPE */
    /* =============================================================== */
    static class MPrimitiveType extends MType implements PrimitiveType {

        MPrimitiveType(MVM vm, int type, String name) {
            super(vm, type, name);
        }
    }

    int type;

    String name;

    public MType(MVM vm, int type, String name) {
        super(vm);
        this.type = type;
        this.name = name;
    }

    public boolean isArray() {
        return this.type == Type.ARRAY;
    }

    public boolean isPrimitive() {
        return this.type != Type.CLASS && this.type != Type.ARRAY;
    }

    @Override
    public CodeWriter toCode(CodeWriter out) {
        out.queue(this.comment);
        boolean foundInImportedTypes = isImportedType(out);
        if (foundInImportedTypes) {
            out.write(name.substring(name.lastIndexOf('.') + 1));
        } else {
            out.write(this.name);
        }
        return out;
    }

    private boolean isImportedType(CodeWriter out) {
        if (this.name.indexOf('.') > 0) {
            if (name.startsWith(JAVA_LANG_PACKAGE)) {
                return true;
            }
            if (out.getCompilationUnit() != null) {
                for (Import importedType : out.getCompilationUnit().getImports()) {
                    if (importedType.getName().equals(this.name)) {
                        return true;
                    }
                    if (importedType.getName().endsWith(".*")) {
                        String packageName = name.substring(0, JAVA_LANG_PACKAGE.length());
                        if ((packageName + "*").equals(importedType.getName())) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    @Override
    public String toString() {
        return this.name;
    }

    public int type() {
        return this.type;
    }
}
